########################################################################
#####Pedro Ivo Siqueira Nepomuceno - 8941321 ###########################
#####pedro.ivo.nepomuceno@usp.br             ###########################
########################################################################

import random
import numpy as np
import time
import itertools
import math
import gurobipy as gb
from time import gmtime, strftime

dbSize = 0
db = []

def exportValues(queries, answers, noises,flMax):
	global db
	fileName = 'test_' + str(flMax) + "_" + strftime("%Y%m%d_%H%M%S", gmtime())
	with open(fileName+'header.txt','w') as logFile:		
		logFile.write(';'.join(str(d) for d in db) + '\n')

	with open(fileName+'queries.txt','w') as logFile:					
		for query in queries:
			logFile.write(';'.join(str(q) for q in query) + '\n')

	with open(fileName+'answers.txt','w') as logFile:					
		for answer in answers:
			logFile.write(str(answer) + '\n')

	with open(fileName+'noises.txt','w') as logFile:					
		for noise in noises:
			logFile.write(str(noise) + '\n')


def getNoise(noiseLimit):
	noiseLimit = int(math.floor(noiseLimit))
	return(random.randint(-noiseLimit,noiseLimit))

def answerQuery(query, noise): 		
	ans = sum(db[list(query)])
	return max(ans+noise ,0)


#Generate one random query
def randomQuery():
	nqueries = random.randint(1,dbSize) #Number of rows that will be selected

	q = set()
	for i in range(nqueries):
		q.add(random.randint(0,dbSize-1))

	return q

#Generate all subsets of a set
def allSubsets(wset):
	x = len(wset)
	masks = [1 << i for i in range(x)]
	for i in range(1 << x):
		yield [ss for mask, ss in zip(masks, wset) if i & mask]



def exponential():
	queries = list(allSubsets(range(dbSize)))##Obtendo todas as consultas possiveis			

	answers = []
	for query in queries:
		answers.append(answerQuery(query,dbSize-4))
	
	possibilities = list(allSubsets(range(dbSize)))	
	
	inicio = time.time()
	for possibility in possibilities:		
		cracked = True
		for i in range(len(queries)):
			if abs(len(set(possibility).intersection(set(queries[i]))) - answers[i]) > (dbSize-4):
				cracked = False
				break
		if cracked:
			dec = []
			for i in range(dbSize):
				if i in possibility:
					dec.append(1)
				else:
					dec.append(0)

			fim = time.time()
			print 'Database decrypted!! Tempo de execucao:',str(fim-inicio )
			return

def lpn():
	t = int(math.floor(dbSize*math.log(dbSize,2)))
	

	lp = gb.Model("DataBase")
	variables = lp.addVars([i for i in range(dbSize)], vtype = gb.GRB.CONTINUOUS) #,vtype = gb.GRB.BINARY)  Tipo deve ser continuo para suportar limites com diferenca

	obj = gb.LinExpr()
	for v in variables:
		obj += variables[v]
		lp.addConstr(variables[v] <= 1)
		lp.addConstr(variables[v] >= 0)
	

	lp.setObjective(obj,gb.GRB.MINIMIZE)

	queries = []
	noises = []
	answers = []
	for i in range(t):
		query = randomQuery()
		queries.append(query)
		noiseLimit = pow(dbSize,1.0/3.0)		
		noise = getNoise(noiseLimit)		
		noises.append(noise)

		ans = answerQuery(query, noise)
		answers.append(ans)


		cons = gb.LinExpr()
		for c in query:
			cons += variables[c]

		cons1 = cons <= ans + noiseLimit		
		cons2 = cons >= ans - noiseLimit		

		lp.addConstr(cons1)
		lp.addConstr(cons2)		

	lp.optimize()
	print 'Database Decrypted!!!'
	# resp = []
	# for c in lp.getVars():
	# 	if c.x > 0.5:
	# 		resp.append(int(1))
	# 	else:
	# 		resp.append(int(0))


	if np.array_equal(db,[int(round(c.x)) for c in lp.getVars()]):
		print 'Decryption Succeded for Minimization'
		lp.setObjective(obj,gb.GRB.MAXIMIZE)
		lp.optimize()

		if np.array_equal(db,[int(round(c.x)) for c in lp.getVars()]):
			print 'Decryption Succeded for Maximization'
		else:
			print 'Decryption Failed for Maximization'
			exportValues(queries,answers,noises,1)

	else:
		print 'Decryption Failed fo Minimization'
		exportValues(queries,answers,noises,0)




def main():
	global dbSize
	global db

	dbSize = input("Database Size:")
	
	repetitions = input("Repeat times:")

	for i in range(dbSize):
		db.append(random.randint(0,1))

	db = np.array(db)

	workingMode = input("Working Mode: (0 - Exponential / 1 - Polinomial:")

	print 'Real DataBase:',str(db)

	for i in range(repetitions):
		print 'STARTING ',str(i), ' TEST'
		if (workingMode == 0):
			exponential()	
		else:
			lpn()




if __name__ == "__main__":
    main()
